import click
import pandas as pd
import requests
import json
import os
import sys
from openpyxl import Workbook
from carpalrouting.grouping import LocationGrouping, DriverGrouping
from carpalrouting.optimize import Routing
from carpalrouting.models import Driver, Location, Coordinates, GeoArea

@click.command()
@click.option('--file', help='The absolute path of setting.json file')
def cli(file):
    if not file:
        print('Error. No schema file provided.')
    else:
        try:
            with open(file) as setting_file:
                print('Loading setting file...')
                settings = json.load(setting_file)

                print('Reading driver data...')
                drivers = settings.get('drivers')

                print('Reading location data...')
                df = pd.read_excel(settings.get('inputFile'))
                locations = {}
                for index, row in df.iterrows():
                    pickup_address = row['pickup_location_address']
                    pickup_window = tuple(map_time_to_seconds(row['pickup_time_window']))

                    if pickup_address in locations:                
                        new_item = {'address': row['delivery_address'],
                                    'coordinates': map_address_to_coordinates(row['delivery_address'], settings.get('googleAPIKey')),
                                    'delivery_window': map_time_to_seconds(row['delivery_time_window']),
                                    'capacity': row['total_weight'],
                                    'order_time': 0}
                        if pickup_window in locations[pickup_address]:             
                            locations[pickup_address][pickup_window].append(new_item)
                        else:
                            locations[pickup_address][pickup_window] = [{
                                'address': pickup_address,
                                'coordinates': map_address_to_coordinates(pickup_address, settings.get('googleAPIKey')),
                                'delivery_window': map_time_to_seconds(row['pickup_time_window']),
                                'capacity': 0,
                                'order_time': 0
                            }]
                            locations[pickup_address][pickup_window].append(new_item)
                    else:
                        locations[pickup_address] = {}

                        locations[pickup_address][pickup_window] = [{
                            'address': pickup_address,
                            'coordinates': map_address_to_coordinates(pickup_address, settings.get('googleAPIKey')),
                            'delivery_window': map_time_to_seconds(row['pickup_time_window']),
                            'capacity': 0,
                            'order_time': 0
                        }]

                        locations[pickup_address][pickup_window].append({
                            'address': row['delivery_address'],
                            'coordinates': map_address_to_coordinates(row['delivery_address'], settings.get('googleAPIKey')),
                            'delivery_window': map_time_to_seconds(row['delivery_time_window']),
                            'capacity':row['total_weight'],
                            'order_time': 0
                        })

                schedules = DriverGrouping().group_drivers_by_time_slot(drivers=[Driver(id=item.get('id'),
                                                                                        time_slots=item.get('time_slots'),
                                                                                        capacity=item.get('capacity'),
                                                                                        speed=item.get('speed')) for item in drivers])
                                                    
                data = LocationGrouping().group_locations(locations, schedules)

                print('Start route optimization...')
                book = Workbook()
                sheet = book.active
                
                for pickup_address, pickup_windows in data.items():
                    for pickup_window, obj in pickup_windows.items():
                        locs = obj.get('locations')
                        driver_ids = obj.get('driver_ids')
                        while locs and driver_ids:
                            driver = None 
                            for item in drivers:
                                if item.get('id') == driver_ids[0]:
                                    driver = item
                                    break

                            #find the indices for all filtered locations in data object
                            filtered_indices = []
                            for k, v in enumerate(locs):
                                if v.capacity <= driver.get('capacity'):
                                    filtered_indices.append(k)
                                    
                            indices = Routing(locations=[loc for loc in locs if loc.capacity <= driver.get('capacity')],
                                              num_vehicles=len(locs)).generate_routes(vehicle_capacity=driver.get('capacity'),
                                                                                      speed=driver.get('speed'),
                                                                                      service_time_unit=settings['serviceTime'])
                            if len(indices) == 0:
                            
                                break
                            else:
                                #remove driver from available driver list only if a route has been built
                                driver_ids.pop(0)                            

                                routes = []
                                for item in indices[0]:
                                    routes.append(locs[filtered_indices[item]])
                                    if filtered_indices[item] != 0:
                                        locs[filtered_indices[item]] = None

                                locs = list(filter(lambda x: x is not None, locs))
                                sheet.append(tuple([item.address for item in routes]))
                                                                
                        obj['locations'] = locs if len(locs) > 1 else [] 
                book.save(settings.get('outputFile'))

                if settings['publicDriver']['required']:
                    print('Using public drivers...')

                    for pickup_address, pickup_windows in data.items():
                        for pickup_window, obj in pickup_windows.items():
                            locs = obj.get('locations')

                            if locs:
                                indices = Routing(locations=locs,
                                                  num_vehicles=len(locs)).generate_routes(vehicle_capacity=settings['publicDriver']['specification']['capacity'],
                                                                                          speed=settings['publicDriver']['specification']['speed'],
                                                                                          service_time_unit=settings['serviceTime'])
                                if len(indices) == 0:                    
                                    break
                                else:                                                
                                    for idx_list in indices:
                                        routes = []

                                        for item in idx_list:
                                            routes.append(locs[item])
                                            if item != 0:
                                                locs[item] = None

                                        sheet.append(tuple([item.address for item in routes]))
                                        
                                    locs = list(filter(lambda x: x is not None, locs))
                                    obj['locations'] = locs if len(locs) > 1 else []
                    book.save(settings.get('outputFile'))
        except Exception as e:
            print('Error. Unable to load the setting file.')
            sys.exit()

def map_address_to_coordinates(address, key):
    response = requests.get('https://maps.googleapis.com/maps/api/geocode/json?address={}&key={}'.format(address, key))
    content = json.loads(response.__dict__.get('_content'))
    if content.get('results'):
        loc = content.get('results')[0].get('geometry').get('location')
        return [loc.get('lat'), loc.get('lng')]
    else:
        return [0, 0]

def map_time_to_seconds(time):
    window = time.split('-')
    start = window[0].split(':')
    end = window[1].split(':')

    return [int(start[0])*3600+int(start[1])*60, int(end[0])*3600+int(end[1])*60]

if __name__ == '__main__':
    cli()


import copy
from .models import Driver, Location, Wave, Coordinates

class DriverGrouping(object):
    def __init__(self):
        pass

    def group(self, drivers, locations, start_buffer_time=0, end_buffer_time=0):
        #Verifying data type        
        illegal_driver_data = [item for item in drivers if not isinstance(item, Driver)]
        illegal_location_data = [item for item in locations if not isinstance(item, Location)]
        
        if illegal_driver_data:
            raise TypeError('All element in the list should have the same type Driver.')

        if illegal_location_data:
            raise TypeError('All element in the list should have the same type Location.')

        result = {}

        for driver in drivers:
            result[driver.id] = {}
            for time_slot in driver.time_slots:
                result[driver.id][(time_slot[0], time_slot[1])] = []
                for idx, val in enumerate(locations):
                    if not val:
                        continue
                    elif (time_slot[0] + start_buffer_time) < val.delivery_window[0] and \
                        val.delivery_window[1]< (time_slot[1] - end_buffer_time) and \
                        self.if_loc_capacity_within(val.capacity, driver.capacity):
                        result[driver.id][(time_slot[0], time_slot[1])].append(val)
                        locations[idx] = None
                    
        result['unprocessed'] = [item for item in locations if item is not None]
        return result

    def group_drivers_by_time_slot(self, drivers):
        #Verifying data type        
        illegal_driver_data = [item for item in drivers if not isinstance(item, Driver)]
        
        if illegal_driver_data:
            raise TypeError('All element in the list should have the same type Driver.')

        time_slots = {}

        for driver in drivers:
            for time_slot in driver.time_slots:
                if (time_slot[0], time_slot[1]) in time_slots:
                    if time_slots[(time_slot[0], time_slot[1])]:
                        time_slots[(time_slot[0], time_slot[1])].append(driver)
                    else:
                        time_slots[(time_slot[0], time_slot[1])] = [driver]
                else:
                    time_slots[(time_slot[0], time_slot[1])] = [driver]
        return time_slots

    def if_loc_capacity_within(self, location_weight, driver_capacity):            
        return driver_capacity > location_weight

class LocationGrouping(object):
    def __init__(self):
        pass

    def _group_locations_by_time_slot(self, locations, time_slot, threshhold=0):
        """
        Return locations that not included in the given time slot
        """
        result = [[], []]
        for loc in locations:
            if threshhold == 0:
                if time_slot[0] < loc.delivery_window[0] and loc.delivery_window[1] < time_slot[1]:
                    result[0].append(loc)
                else:
                    result[1].append(loc)
            else:
                #TODO: make use of threshhold argument
                if (time_slot[0] < loc.delivery_window[0] < time_slot[1] and loc.delivery_window[0] == (time_slot[0]+time_slot[1])/2) or \
                    (time_slot[0] < loc.delivery_window[1] < time_slot[1] and loc.delivery_window[0] == (time_slot[0]+time_slot[1])/2):
                    result[0].append(loc)
                else:
                    result[1].append(loc)
        return result

    def group_locations(self, locations, schedules):
        """
        This method requires the locations to be grouped by pickup locations

        Sample processed:
        {'pickup1': {(12000, 18000) : {(43140, 79200): [<carpalrouting.models.Location object at 0x110b62be0>], 
                     (0, 43140): [<carpalrouting.models.Location object at 0x1127d66d8>, <carpalrouting.models.Location object at 0x112a15748>]}}, 
         'pickup2': {(12000, 18000) : {(28800, 79200): [<carpalrouting.models.Location object at 0x110778320>, <carpalrouting.models.Location object at 0x112f0ecf8>, <carpalrouting.models.Location object at 0x112f0ee48>]}}, 
         'pickup3': {(12000, 18000) : {(28800, 79200): [<carpalrouting.models.Location object at 0x112a15dd8>]}}}

        Sample unprocessed: 
        {'pickup1': [], 'pickup2': [], 'pickup3': [<carpalrouting.models.Location object at 0x110778390>]}
        """
        result = {}
        copied_schedules = copy.deepcopy(schedules)

        for pickup, pickup_windows in locations.items():
            result[pickup] = {}
            for pickup_window, dropoff_list in pickup_windows.items():                             
                result[pickup][pickup_window] = {'locations': [Location(coordinates=Coordinates(item.get('coordinates')[0], item.get('coordinates')[1]),
                                                                        address=item.get('address'),
                                                                        delivery_window=item.get('delivery_window'),
                                                                        capacity=item.get('capacity', 1),
                                                                        ref=item.get('ref', '')) for item in dropoff_list], 'driver_ids': []}
                        
                total_capacity = 0

                for item in result[pickup][pickup_window]['locations']:
                    total_capacity += item.capacity

                for time_slot, driver_list in copied_schedules.items():
                    if driver_list:
                        if time_slot[0] <= pickup_window[0] and time_slot[1] >= pickup_window[1]:
                            driver = driver_list.pop(0)
                            result[pickup][pickup_window]['driver_ids'].append(driver.id)
                            total_capacity -= driver.capacity  
                                                    
                            while total_capacity > 0 and driver_list:                                                               
                                driver = driver_list.pop(0)
                                result[pickup][pickup_window]['driver_ids'].append(driver.id)
                                total_capacity -= driver.capacity
            
                    
        return result
        
class WaveGrouping(object):
    def __init__(self):
        pass

    def group(self, waves, locations, buffer_time=0):
        illegal_wave_data = [item for item in waves if not isinstance(item, Wave)]
        illegal_location_data = [item for item in locations if not isinstance(item, Location)]

        if illegal_wave_data:
            raise TypeError('All element in the list should have the same type Wave.')

        if illegal_location_data:
            raise TypeError('All element in the list should have the same type Location.')

        result = {}
        for wave in waves:
            result[wave.cutoff_time] = [locations.pop(idx) for idx, val in enumerate(locations) if \
                val.order_time + buffer_time < wave.cutoff_time]

        if len(locations) > 0:
            result['unprocessed'] = locations
            
        return result


from ortools.constraint_solver import pywrapcp
from math import sin, cos, sqrt, atan2, asin, radians

class CreateDistanceCallback(object):
    """Create callback to calculate distances between points."""

    def __init__(self, locations):
        """Initialize distance array."""
        size = len(locations)
        self.matrix = {}

        for from_node in range(size):
            self.matrix[from_node] = {}
            for to_node in range(size):
                x1 = locations[from_node][0]
                y1 = locations[from_node][1]
                x2 = locations[to_node][0]
                y2 = locations[to_node][1]
                self.matrix[from_node][to_node] = CreateDistanceCallback.distance(x1, y1, x2, y2)

    def distance_callback(self, from_node, to_node):
        return int(self.matrix[from_node][to_node])

    @staticmethod
    def distance(lat1, long1, lat2, long2):
        lat1, long1, lat2, long2 = map(radians, [lat1, long1, lat2, long2])

        # haversine formula 
        dlon = long2 - long1 
        dlat = lat2 - lat1 

        a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
        c = 2 * asin(sqrt(a)) 
        r = 6371 # Radius of earth in kilometers. Use 3956 for miles
        return c * r

# Demand callback
class CreateDemandCallback(object):
    """Create callback to get demands at each location."""

    def __init__(self, demands):
        self.matrix = demands

    def demand_callback(self, from_node, to_node):
        return self.matrix[from_node]

# Service time (proportional to demand) callback.
class CreateServiceTimeCallback(object):
    """Create callback to get time windows at each location."""

    def __init__(self, demands, time_per_demand_unit):
        self.matrix = demands
        self.time_per_demand_unit = time_per_demand_unit

    def ServiceTime(self, from_node, to_node):
        try:
            result = int(self.matrix[from_node] * self.time_per_demand_unit)
            return result
        except Exception as e:
            import traceback
            traceback.print_exc()
            print(self.matrix)
            print(from_node)

# Create the travel time callback (equals distance divided by speed).
class CreateTravelTimeCallback(object):
    """Create callback to get travel times between locations."""
    def __init__(self, dist_callback, speed):
        self.dist_callback = dist_callback
        self.speed = speed

    def TravelTime(self, from_node, to_node):
        travel_time = self.dist_callback(from_node, to_node) / self.speed
        return int(travel_time)

# Create total_time callback (equals service time plus travel time).
class CreateTotalTimeCallback(object):
    """Create callback to get total times between locations."""

    def __init__(self, service_time_callback, travel_time_callback):
        self.service_time_callback = service_time_callback
        self.travel_time_callback = travel_time_callback

    def TotalTime(self, from_node, to_node):
        service_time = self.service_time_callback(from_node, to_node)
        travel_time = self.travel_time_callback(from_node, to_node)
        return service_time + travel_time    

class Routing(object):
    def __init__(self, locations, num_vehicles=1, depot=0):
        """
        depot: the time spent on departure for things like loading stuff etc...
        """       
        self._locations = [item.coordinates for item in locations]
        self._num_locations = len(locations)
        self._max_weights = [item.capacity for item in locations]
        self._num_vehicles = num_vehicles
        self._depot = depot
        self._delivery_windows= [item.delivery_window for item in locations]

        # The number of nodes of the VRP is num_locations.
        # Nodes are indexed from 0 to num_locations - 1. By default the starting of
        # a route is node 0.
        self._routing = pywrapcp.RoutingModel(self._num_locations, self._num_vehicles, self._depot)
        self._search_parameters = pywrapcp.RoutingModel.DefaultSearchParameters()
        self._search_parameters.time_limit_ms = 5000
        self._search_parameters.solution_limit = 1

    def find_nearest_pickup_location(self, pickup_locations, first_location_in_route):
        try:
            nearest = 0
            distance = CreateDistanceCallback.distance(pickup_locations[nearest][0], 
                                                       pickup_locations[nearest][1],
                                                       first_location_in_route[0],
                                                       first_location_in_route[1])
                                                       
            for index, value in enumerate(pickup_locations):
                dist = CreateDistanceCallback.distance(value[0], 
                                                       value[1],
                                                       first_location_in_route[0],
                                                       first_location_in_route[1])
                if  dist < distance:
                    nearest = index
                    distance = dist
                else:
                    continue
            return nearest
        except Exception as e:
            raise e

    def generate_routes(self, vehicle_capacity, speed, service_time_unit=1, working_hours=24):
        try:
            # Callbacks to the distance function and travel time functions here.
            dist_between_locations = CreateDistanceCallback(self._locations)
            dist_callback = dist_between_locations.distance_callback

            self._routing.SetArcCostEvaluatorOfAllVehicles(dist_callback)
            demands_at_locations = CreateDemandCallback(self._max_weights)
            demands_callback = demands_at_locations.demand_callback

            # Adding capacity dimension constraints.
            slack_max = 0;
            fix_start_cumul_to_zero = True

            self._routing.AddDimension(demands_callback,
                                       slack_max,
                                       vehicle_capacity,
                                       fix_start_cumul_to_zero,
                                       "Capacity")
            # Add time dimension.
            time_per_demand_unit = service_time_unit
            horizon = working_hours * 3600
            time = "Time"

            service_times = CreateServiceTimeCallback(self._max_weights,
                                                    time_per_demand_unit)
            service_time_callback = service_times.ServiceTime

            travel_times = CreateTravelTimeCallback(dist_callback, speed)
            travel_time_callback = travel_times.TravelTime

            total_times = CreateTotalTimeCallback(service_time_callback, travel_time_callback)
            total_time_callback = total_times.TotalTime

            self._routing.AddDimension(total_time_callback,  # total time function callback
                                       horizon,
                                       horizon,
                                       fix_start_cumul_to_zero,
                                       time)

            # Add time window constraints.
            time_dimension = self._routing.GetDimensionOrDie(time)
            for location in range(1, self._num_locations):
                start = self._delivery_windows[location][0]
                end = self._delivery_windows[location][1]
                time_dimension.CumulVar(location).SetRange(start, end)

            # Solve displays a solution if any.
            assignment = self._routing.SolveWithParameters(self._search_parameters)  

            if assignment:
                capacity_dimension = self._routing.GetDimensionOrDie("Capacity");
                time_dimension = self._routing.GetDimensionOrDie(time);

                possible_indices = []
                used_location = []           

                for vehicle_nbr in range(self._num_vehicles):
                    index = self._routing.Start(vehicle_nbr)
                    
                    # temp_list = []
                    # temp_addr_list = []
                    index_list = []

                    while not self._routing.IsEnd(index):
                        node_index = self._routing.IndexToNode(index)
                        load_var = capacity_dimension.CumulVar(index)
                        time_var = time_dimension.CumulVar(index)
                        
                        # temp_list.append(self._locations[node_index])
                        # temp_addr_list.append(self._addresses[node_index])
                        index_list.append(node_index)
                        used_location.append(self._locations[node_index])
                        index = assignment.Value(self._routing.NextVar(index))
                   
                    node_index = self._routing.IndexToNode(index)
                    load_var = capacity_dimension.CumulVar(index)
                    time_var = time_dimension.CumulVar(index)

                    # temp_list.append(self._locations[node_index])
                    # temp_addr_list.append(self._addresses[node_index])
                    #index_list.append(node_index)

                    #used_location.append(self._locations[node_index])

                    if len(index_list) > 1:
                        # del temp_list[-1]
                        # del temp_addr_list[-1]

                        # possible_routes.append(temp_list)
                        # possible_addresses.append(temp_addr_list)

                        possible_indices.append(index_list)
                # unused = []

                # for item in self._locations:            
                #     if not item in used_location:                                   
                #         unused.append(item)

                return possible_indices
            else:
                return []
        except Exception as e:
            raise e
    


import copy
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
from .models import Coordinates, GeoArea, Location

class GeoTools(object):
    @staticmethod
    def is_point_in_area(edge_coordinates, coordinates):
        polygon = []
        for item in edge_coordinates:
            polygon.append((item[0], item[1]))

        return Polygon(polygon).contains(Point(coordinates[0], coordinates[1]))

    def group_locations_by_time_slot(self, locations, time_slot, threshhold=0):
        """
        Return locations that not included in the given time slot
        """
        result = [[], []]
        for loc in locations:
            if threshhold == 0:
                if time_slot[0] < loc.delivery_window[0] and loc.delivery_window[1] < time_slot[1]:
                    result[0].append(loc)
                else:
                    result[1].append(loc)
            else:
                #TODO: make use of threshhold argument
                if (time_slot[0] < loc.delivery_window[0] < time_slot[1] and loc.delivery_window[0] == (time_slot[0]+time_slot[1])/2) or \
                    (time_slot[0] < loc.delivery_window[1] < time_slot[1] and loc.delivery_window[0] == (time_slot[0]+time_slot[1])/2):
                    result[0].append(loc)
                else:
                    result[1].append(loc)
        return result

    @staticmethod
    def group_locations(edge_coordinates, delivery_locations, pickup_location, schedules=[]):
        """
        This method will group locations by multiple areas with groups of edge locations

        Sample processed:
        {'pickup1': {(43140, 79200): [<carpalrouting.models.Location object at 0x110b62be0>], 
                     (0, 43140): [<carpalrouting.models.Location object at 0x1127d66d8>, <carpalrouting.models.Location object at 0x112a15748>]}, 
         'pickup2': {(28800, 79200): [<carpalrouting.models.Location object at 0x110778320>, <carpalrouting.models.Location object at 0x112f0ecf8>, <carpalrouting.models.Location object at 0x112f0ee48>]}, 
         'pickup3': {(28800, 79200): [<carpalrouting.models.Location object at 0x112a15dd8>]}}

        Sample unprocessed: 
        {'pickup1': [], 'pickup2': [], 'pickup3': [<carpalrouting.models.Location object at 0x110778390>]}

        """
        locations = []        
        area = GeoArea()
        area.edge_locations = [v for k, v in edge_coordinates.items()]

        for item in delivery_locations:
            if GeoTools.is_point_in_area(area.edge_locations, item.get('coordinates')):  
                locations.append(item)

        all_processed = {}
        all_unprocessed = {}
        copied_schedules = copy.deepcopy(schedules)
        
        getTool = GeoTools()
        for pickup, dropoff_list in {pickup_location.get('address'): locations}.items():
            #clone delivery locations of each pick up location
            dropoffs = dropoff_list[:]
            processed_locs = {}   

            #loop all available driver scheduels
            for time_slot, driver_list in copied_schedules.items():
                processed_locs[time_slot] = []  

                # The loop will continue the check 
                # if there are still unprocessed delivery locations and available drivers
                while(dropoffs and driver_list):
                    grp = getTool.group_locations_by_time_slot(
                        locations=[Location(coordinates=Coordinates(item.get('coordinates')[0], item.get('coordinates')[1]),
                                    address=item.get('address'),
                                    delivery_window=item.get('delivery_window'),
                                    capacity=item.get('capacity'),
                                    order_time=item.get('order_time')) for item in dropoffs],
                        time_slot=time_slot)

                    #if locations partially processed, pop a driver from the pool
                    driver_list.pop(0)
                    processed_locs[time_slot].extend(grp[0])

                    dropoffs = [{'coordinates':[item.coordinates[0], item.coordinates[1]],
                                 'address': item.address,
                                 'delivery_window': item.delivery_window,
                                 'capacity': item.capacity,
                                 'order_time': item.order_time} for item in grp[1]]

            all_unprocessed.update({pickup: [Location(coordinates=Coordinates(item.get('coordinates')[0], item.get('coordinates')[1]),
                                                      address=item.get('address'),
                                                      delivery_window=item.get('delivery_window'),
                                                      capacity=item.get('capacity'),
                                                      order_time=item.get('order_time')) for item in dropoffs]})
            all_processed.update({pickup: processed_locs})
        return (all_processed, all_unprocessed)


import json
import pickle

class Driver(object):
    def __init__(self, id, time_slots, capacity, speed, ref=''):
        self._id = id
        self._time_slots = time_slots
        self._capacity = capacity
        self._speed = speed
        self._ref = ref

    @property
    def id(self):
        return self._id

    @property
    def time_slots(self):
        return self._time_slots

    @property
    def capacity(self):
        return self._capacity

    @property
    def speed(self):
        return self._speed
    
    @property
    def ref(self):
        return self._ref

    def to_json(self):
        return {'id': self._id,
                'time_slots': self._time_slots,
                'capacity': self._capacity,
                'speed': self._speed,
                'ref': self._ref}

class Location(object):
    def __init__(self, coordinates, address, delivery_window, capacity, order_time=0, ref=''):
        if not isinstance(coordinates, Coordinates):
            raise TypeError('Coordinates object expected.')

        self._coordinates = coordinates.coordinates
        self._address = address
        self._delivery_window= delivery_window
        self._capacity = capacity
        self._order_time = order_time
        self._ref = ref

    @property
    def coordinates(self):
        return self._coordinates

    @property
    def address(self):
        return self._address

    @property
    def delivery_window(self):
        return self._delivery_window

    @property
    def order_time(self):
        return self._order_time

    @property
    def capacity(self):
        return self._capacity

    @property
    def ref(self):
        return self._ref    

    def to_json(self):
        m, s = divmod(self._delivery_window[0], 60)
        h, m = divmod(m, 60)
        fr = '%02d:%02d' % (h, m)

        m, s = divmod(self._delivery_window[1], 60)
        h, m = divmod(m, 60)
        to = '%02d:%02d' % (h, m)
        return {'coordinates': [self._coordinates[0], self._coordinates[1]],
                'address': self._address,
                'delivery_window': [fr, to],
                'order_time': self._order_time,
                'capacity': self._capacity,
                'ref': self._ref}

class Coordinates(object):
    def __init__(self, latitude, longitude):
        self._latitude = latitude
        self._longitude = longitude

    @property
    def coordinates(self):
        return [self._latitude, self._longitude]

    def to_json(self):
        return {'coordinates': [self._latitude, self._longitude]}

class GeoArea(object):    
    _edge_locations = []
    _locations = []

    @property
    def edge_locations(self):
        return self._edge_locations

    @edge_locations.setter
    def edge_locations(self, edge_locations):
        self._edge_locations = edge_locations

    @property
    def locations(self):
        return self._locations

    @locations.setter
    def locations(self, locations):
        self._locations = locations

    def __str__(self):
        #TODO: make sure this returns 
        return pickle.dumps(self.__dict__)

class Wave(object):
    def __init__(self, cutoff_time):
        self._cutoff_time = cutoff_time

    @property
    def cutoff_time(self):
        return self._cutoff_time

    def __str__(self):
        return self.__dict__



import os, sys
current_path = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, current_path + '/../') 
from carpalrouting.models import Driver, Location, Coordinates
from carpalrouting.grouping import DriverGrouping, LocationGrouping
from carpalrouting.optimize import Routing
from decimal import Decimal
import unittest

class TestDriverLocationGrouping(unittest.TestCase):
    def setUp(self):
        self._drivers = [
            {
                'id': 981,
                'time_slots': [[53000, 61600],[30000, 42000]],
                'capacity': 100,
                'speed': 20
            },{
                'id': 1000,
                'time_slots': [[53000, 61600],[30000, 42000]],
                'capacity': 16,
                'speed': 5
            }]
            
        self._locations = [
            {
                'coordinates': [-33.8815474, 150.9625217],
                'delivery_window': [64800, 72000],
                'capacity': 7.8,
                'order_time': 30660
            },{
                'coordinates': [-33.8322542, 151.0548105],
                'delivery_window': [32400, 39600],
                'capacity': 15.8,
                'order_time': 33600
            },{
                'coordinates': [-33.881716, 150.9136746],
                'delivery_window': [32400, 39600],
                'capacity': 3.8,
                'order_time': 45000
            },{
                'coordinates': [-33.9545249, 150.9406193],
                'delivery_window': [54000, 61200],
                'capacity': 35.5,
                'order_time': 55800
            }]

    def test_driver_schedule_grouping(self):
        self.assertRaises(TypeError, lambda: DriverGrouping().group(drivers=self._drivers,
                                                                    locations=self._locations))

        result = DriverGrouping().group(drivers=[Driver(id=item.get('id'),
                                                        time_slots=item.get('time_slots'),
                                                        capacity=item.get('capacity'),
                                                        speed=item.get('speed')) for item in self._drivers],
                                        locations=[Location(coordinates=Coordinates(item.get('coordinates')[0], item.get('coordinates')[1]),
                                                            address=item.get('address'),
                                                            delivery_window=item.get('delivery_window'),
                                                            capacity=item.get('capacity'),
                                                            order_time=item.get('order_time')) for item in self._locations])
        
        self.assertEqual(len(list(result.items())), 3)

    def test_driver_schedule_multiple_pickup_windows_grouping(self):
        drivers = [{
                'id':1,
                'name': 'Driver 1',
                'time_slots': [[0, 43140], [43140, 79200]],
                'capacity': 500,
                'speed': 50
            },
            {
                'id':2,
                'name': 'Driver 2',
                'time_slots': [[0, 43140]],
                'capacity': 400,
                'speed': 50
            },
            {'id':3,
                'name': 'Driver 3',
                'time_slots': [[0, 43140]],
                'capacity': 500,
                'speed': 50
            },
            {'id':4,
                'name': 'Driver 4',
                'time_slots': [[28800, 43140]],
                'capacity': 500,
                'speed': 50
            },
            {'id':5,
                'name': 'Driver 5',
                'time_slots': [[28800, 79200]],
                'capacity': 500,
                'speed': 50
            },
            {'id':6,
                'name': 'Driver 6',
                'time_slots': [[0, 43140]],
                'capacity': 500,
                'speed': 50
            },
            {'id':7,
                'name': 'Driver 7',
                'time_slots': [[0, 43140], [58800, 80000]],
                'capacity': 500,
                'speed': 50
            },
            {'id':8,
                'name': 'Driver 8',
                'time_slots': [[28800, 79200]],
                'capacity': 500,
                'speed': 50
            },
            {'id':9,
                'name': 'Driver 9',
                'time_slots': [[0, 79200]],
                'capacity': 500,
                'speed': 50
            },
            {'id':10,
                'name': 'Driver 10',
                'time_slots': [[0, 50800], [60400, 89200]],
                'capacity': 800,
                'speed': 60
            }]
            
        locations = {
            (1.3001, 102.0001):
            {
                (12000, 20000):[
                    {
                        'coordinates': [-33.8815474, 150.9625217],
                        'address': 'pick up 1',
                        'delivery_window': [12000, 20000],
                        'capacity': 0,
                        'order_time': 30660
                    },{
                        'coordinates': [-33.8322542, 151.0548105],
                        'address': 'Delivery 1',
                        'delivery_window': [16400, 39600],
                        'capacity': 65.8,
                        'order_time': 33600
                    },{
                        'coordinates': [-33.881716, 150.9136746],
                        'address': 'Delivery 2',
                        'delivery_window': [22400, 39600],
                        'capacity': 43.8,
                        'order_time': 45000
                    }
                ],(52000, 58600):[
                    {
                        'coordinates': [-33.8815431, 151.9234],
                        'address': 'pickup 2',
                        'delivery_window': [54800, 58600],
                        'capacity': 0,
                        'order_time': 30660
                    },{
                        'coordinates': [-32.8311, 152.2341],
                        'address': 'delivery 3',
                        'delivery_window': [60000, 68000],
                        'capacity': 15.8,
                        'order_time': 33600
                    }
                ]
            },
            (1.3002, 102.0002):
            {
                (40000, 44000):[{
                    'coordinates': [-33.9545249, 150.9406193],
                    'address': 'pickup 3',
                    'delivery_window': [40000, 44000],
                    'capacity': 0,
                    'order_time': 55800
                },{
                    'coordinates': [-33.9375904, 150.8350134],
                    'address': 'delivery 4',
                    'delivery_window': [44000, 61200],
                    'capacity': 15.5,
                    'order_time': 35800
                },{
                    'coordinates': [-33.8800401, 151.1104426],
                    'address': 'delivery 5',
                    'delivery_window': [44000, 61200],
                    'capacity': 25,
                    'order_time': 45800
                }]
            },
            (1.3003, 102.0003):
                {(20000, 26800):[{
                    'coordinates': [-33.7171208, 151.0952565],
                    'address': 'pickup 4',
                    'delivery_window': [20000, 26800],
                    'capacity': 0,
                    'order_time': 55800
                },{
                    'coordinates': [-33.9545249, 150.9406193],
                    'address': 'delivery 6',
                    'delivery_window': [36000, 41200],
                    'capacity': 35.5,
                    'order_time': 55800
                }]
            }
        }
        
        schedules = DriverGrouping().group_drivers_by_time_slot(drivers=[Driver(id=item.get('id'),
                                                                                time_slots=item.get('time_slots'),
                                                                                capacity=item.get('capacity'),
                                                                                speed=item.get('speed')) for item in drivers])

        data = LocationGrouping().group_locations(locations, schedules)

        #self.assertEqual(len(all_routes), 4)



import os, sys
current_path = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, current_path + '/../') 
import unittest
from carpalrouting.optimize import Routing
from carpalrouting.models import Driver, Location, Coordinates
from carpalrouting.grouping import DriverGrouping, LocationGrouping
import time


class TestOptimizer(unittest.TestCase):   
    def test_optimization_with_predefined_data(self):
        drivers = [{'id':1,'time_slots': [[32400, 64800]], 'capacity': 3, 'speed': 30},            
                   {'id':2,'time_slots': [[32400, 64800]], 'capacity': 3, 'speed': 30}, 
                   {'id':3,'time_slots': [[32400, 64800]], 'capacity': 3, 'speed': 30}, 
                   {'id':4,'time_slots': [[32400, 64800]], 'capacity': 3, 'speed': 30}, 
                   {'id':5,'time_slots': [[32400, 64800]], 'capacity': 3, 'speed': 30},
                   {'id':6,'time_slots': [[32400, 64800]], 'capacity': 15, 'speed': 30}, 
                   {'id':7,'time_slots': [[32400, 64800]], 'capacity': 15, 'speed': 30}]
           
        locations = {'Fishermall, Diliman, Quezon City': {
                    (50400, 54000): [
                        {'address': 'Fishermall, Diliman, Quezon City', 'coordinates': [14.6336204, 121.0193792], 'delivery_window': [50400, 54000], 'capacity': 0, 'order_time': 0}, 
                        {'address': '81F Madasalin St Teachers Village East Diliman, Quezon City', 'coordinates': [14.6410582, 121.0591618], 'delivery_window': [57600, 64800], 'capacity': 2, 'order_time': 0}], 
                    (32400, 36000): [
                        {'address': 'Fishermall, Diliman, Quezon City', 'coordinates': [14.6336204, 121.0193792], 'delivery_window': [32400, 36000], 'capacity': 0, 'order_time': 0}, 
                        {'address': 'Pacific Rim Cor. Commerce Ave., Filinvest City, Alabang, Muntinlupa City, 1781 Metro Manila, Philippines', 'coordinates': [14.4200106, 121.033259], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 
                        {'address': '4 Kamias Rd, Diliman, Quezon City, Metro Manila', 'coordinates': [14.6310151, 121.0467182], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 		
                        {'address': '30 Humility St Multinational Village, Paranaque', 'coordinates': [14.4804117, 121.0048727], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 		
                        {'address': '7 Marymount La Vista, Quezon City', 'coordinates': [14.6297431, 121.0907328], 'delivery_window': [39600, 46800], 'capacity': 2, 'order_time': 0}, 
                        {'address': 'Blk 6 lot 13 kalayaan village, brgy. 201, Pasay', 'coordinates': [14.507602, 121.029081], 'delivery_window': [39600, 46800], 'capacity': 2, 'order_time': 0}, 
                        {'address': 'Barangay 34, Maypajo, Caloocan, Metro Manila, Philippines', 'coordinates': [14.6367721, 120.9729332], 'delivery_window': [39600, 46800], 'capacity': 2, 'order_time': 0}, 		
                        {'address': '2611 Taft Ave, Barangay 68, Pasay City', 'coordinates':[14.5783882, 120.986897], 'delivery_window': [39600, 46800], 'capacity': 2, 'order_time': 0}, 
                        {'address': '1838 Yakal St, Manila, 237', 'coordinates': [14.6156181, 120.9790745], 'delivery_window': [39600, 46800], 'capacity': 4, 'order_time': 0}, 
                        {'address': '1838 Yakal St 237, Manila', 'coordinates': [14.6156181, 120.9790745], 'delivery_window': [39600, 46800], 'capacity': 4, 'order_time': 0}, 
                        {'address': '18 Molave St., Marietta-Romeo Village, Brgy Sta Lucia, Pasig', 'coordinates': [14.5815801, 121.1017699], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 
                        {'address': '86E Cotabato St. Bago Bantay Alicia, Quezon City', 'coordinates': [14.6619823, 121.0235792], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 		
                        {'address': 'BLK 10 Lot 7 Cor Chesa St Model Community, Bgy 120, Zone 009 Tondo, Manila, Philippines', 'coordinates': [14.6205654, 120.9637766], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 
                        {'address': 'Le Gran Condominium Eisenhower St. Greenhills, San Juan', 'coordinates': [14.607297, 121.048476], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 		
                        {'address': '1838 Yakal St 237, Manila', 'coordinates': [14.6156181, 120.9790745], 'delivery_window': [39600, 46800], 'capacity': 3, 'order_time': 0}, 
                        {'address': '1838 Yakal St 237, Manila', 'coordinates': [14.6156181, 120.9790745], 'delivery_window': [39600, 46800], 'capacity': 4, 'order_time': 0}, 
                        {'address': '1320 CP Garcia St Moriones, Bgy 123, Zone 009 Tondo, Manila', 'coordinates': [14.6205654, 120.9659653], 'delivery_window': [39600, 46800], 'capacity': 2, 'order_time': 0},  
                        {'address': 'Fairways Towers 5th Avenue corner McKinley Road, Taguig\nUnit N 15A/15/Fairways Towers', 'coordinates': [14.545619, 121.0454128], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 
                        {'address': '1310 C. P. Garcia Street, Tondo, Manila, Metro Manila, Philippines', 'coordinates': [14.6105919, 120.9617521], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 
                        {'address': '1320 CP Garcia St Moriones, Bgy 123, Zone 009 Tondo, Manila', 'coordinates': [14.6205654, 120.9659653], 'delivery_window': [39600, 46800], 'capacity': 2, 'order_time': 0}, 
                        {'address': '2176 Jesus St pandacan manila, barangay 848/6', 'coordinates': [14.5926865, 121.0044297], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 
                        {'address': '156 Session Road Quezon City Batasan Hills', 'coordinates': [14.6797936, 121.1032391], 'delivery_window': [39600, 46800], 'capacity': 1, 'order_time': 0}, 
                        {'address': '1838 Yakal St 237, Manila', 'coordinates': [14.6156181, 120.9790745], 'delivery_window': [39600, 46800], 'capacity': 5, 'order_time': 0}], 
                    (39600, 43200): [
                        {'address': 'Fishermall, Diliman, Quezon City', 'coordinates': [14.6336204, 121.0193792], 'delivery_window': [39600, 43200], 'capacity': 0, 'order_time': 0}, 
                        {'address': '4H Anaheim 2, California Garden Square, D.M. Guevara St., Highway Hills, Mandaluyong', 'coordinates': [14.5792302, 121.0452694], 'delivery_window': [46800, 57600], 'capacity': 2, 'order_time': 0}, 
                        {'address': '45 Lebanon Street Don Bosco, Paranaque','coordinates': [14.4786641, 121.032481], 'delivery_window': [46800, 57600], 'capacity': 1, 'order_time': 0}], 
                    (57600, 61200): [
                        {'address': 'Fishermall, Diliman, Quezon City', 'coordinates': [14.6336204, 121.0193792], 'delivery_window': [57600, 61200], 'capacity': 0, 'order_time': 0}, 
                        {'address': '940, MaligayaStreet, Manila', 'coordinates': [14.5665457, 120.9952634], 'delivery_window': [64800, 72000], 'capacity': 1, 'order_time': 0}, 
                        {'address': '1838 Yakal St 237, Manila', 'coordinates': [14.6156181, 120.9790745], 'delivery_window': [64800, 72000], 'capacity': 5, 'order_time': 0}, 
                        {'address': '1838 Yakal St 237, Manila', 'coordinates': [14.6156181, 120.9790745], 'delivery_window': [64800, 72000], 'capacity': 5, 'order_time': 0}]}}
   
        schedules = DriverGrouping().group_drivers_by_time_slot(drivers=[Driver(id=item.get('id'),
                                                                            time_slots=item.get('time_slots'),
                                                                            capacity=item.get('capacity'),
                                                                            speed=item.get('speed')) for item in drivers])
        data = LocationGrouping().group_locations(locations, schedules)

        for pickup_address, pickup_windows in data.items():
            for pickup_window, obj in pickup_windows.items():
                locs = obj.get('locations')
                driver_ids = obj.get('driver_ids')
                while locs and driver_ids:
                    driver = None 
                    for item in drivers:
                        if item.get('id') == driver_ids[0]:
                            driver = item
                            break

                    filtered_indices = []
                    for k, v in enumerate(locs):
                        if v.capacity <= driver.get('capacity'):
                            filtered_indices.append(k)
                            
                    indices = Routing(locations=[loc for loc in locs if loc.capacity <= driver.get('capacity')],
                                      num_vehicles=len(locs)).generate_routes(vehicle_capacity=driver.get('capacity'),
                                                                              speed=driver.get('speed'),
                                                                              service_time_unit=1800)
                    if len(indices) == 0:
                       
                        break
                    else:
                        #remove driver from available driver list only if a route has been built
                        driver_ids.pop(0)                            

                        routes = []
                        for item in indices[0]:
                            routes.append(locs[filtered_indices[item]])
                            if filtered_indices[item] != 0:
                                locs[filtered_indices[item]] = None

                        locs = list(filter(lambda x: x is not None, locs))

                        print([route.capacity for route in routes])
                        print([route.address for route in routes])

                obj['locations'] = locs if len(locs) > 1 else []

        print('Second round...')
        for pickup_address, pickup_windows in data.items():
            for pickup_window, obj in pickup_windows.items():
                locs = obj.get('locations')
                if locs:
                    indices = Routing(locations=locs,
                                      num_vehicles=len(locs)).generate_routes(
                                                                            vehicle_capacity=30,
                                                                            speed=30,
                                                                            service_time_unit=1800)

                    if len(indices) == 0:                    
                        break
                    else:                                                
                        for idx_list in indices:
                            routes = []

                            for item in idx_list:
                                routes.append(locs[item])
                                if item != 0:
                                    locs[item] = None

                            print([route.capacity for route in routes])
                            print([route.address for route in routes])
                            
                        locs = list(filter(lambda x: x is not None, locs))
                        obj['locations'] = locs if len(locs) > 1 else []
        print(data)


from setuptools import setup, find_packages

setup(
    name='CarpalRouting',
    keywords=['routing optimization', 'last mile'],
    description='A routing optimization engine.',
    py_modules=['manage'],
    install_requires=['ortools', 'pandas', 'requests', 'coverage', 'shapely', 'xlrd', 'Click', 'openpyxl'],
    entry_points='''
        [console_scripts]
        routing=manage:cli
    ''',
    license='Apache License 2.0',
    version='0.0.0.35',
    author='CarPal Fleet',
    author_email='nick@carpal.me',
    packages=find_packages(),
    platforms='any',
)